﻿#include "snake_geometry.h"
#include <mapbox/polylabel.hpp>
#include <mapbox/geometry.hpp>

#include <boost/geometry.hpp>
#include <boost/geometry/geometries/polygon.hpp>
#include <boost/geometry/geometries/adapted/boost_tuple.hpp>


#include <GeographicLib/Geocentric.hpp>
#include <GeographicLib/LocalCartesian.hpp>

using namespace mapbox;
using namespace snake_geometry;
using namespace std;

namespace bg = bg;
namespace trans = bg::strategy::transform;

BOOST_GEOMETRY_REGISTER_BOOST_TUPLE_CS(cs::cartesian)


namespace snake_geometry {

void toENU(const GeoPoint3D &WGS84Reference, const GeoPoint3D &WGS84Position, Point3D &ENUPosition)
{
    GeographicLib::Geocentric earth(GeographicLib::Constants::WGS84_a(), GeographicLib::Constants::WGS84_f());
    GeographicLib::LocalCartesian proj(WGS84Reference[0], WGS84Reference[1], WGS84Reference[2], earth);

    proj.Forward(WGS84Position[0], WGS84Position[1], WGS84Position[2], ENUPosition[0], ENUPosition[1], ENUPosition[2]);
}

void fromENU(const Point3D &WGS84Reference, const Point3D &CartesianPosition, GeoPoint3D &GeoPosition)
{
    GeographicLib::Geocentric earth(GeographicLib::Constants::WGS84_a(), GeographicLib::Constants::WGS84_f());
    GeographicLib::LocalCartesian proj(WGS84Reference[0], WGS84Reference[1], WGS84Reference[2], earth);

    proj.Reverse(CartesianPosition[0], CartesianPosition[1], CartesianPosition[2], GeoPosition[0], GeoPosition[1], GeoPosition[2]);
}

void polygonCenter(const BoostPolygon &polygon, BoostPoint &center)
{
   if (polygon.outer().empty())
       return;
   geometry::polygon<double> p;
   geometry::linear_ring<double> lr1;
   for (size_t i = 0; i < polygon.outer().size(); ++i) {
       geometry::point<double> vertex(polygon.outer()[i].get<0>(), polygon.outer()[i].get<1>());
       lr1.push_back(vertex);
   }
   p.push_back(lr1);
   geometry::point<double> c = polylabel(p);

   center.set<0>(c.x);
   center.set<1>(c.y);
}

void minimalBoundingBox(const BoostPolygon &polygon, min_bbox_rt &minBBox)
{
    /*
    Find the minimum-area bounding box of a set of 2D points

    The input is a 2D convex hull, in an Nx2 numpy array of x-y co-ordinates.
    The first and last points points must be the same, making a closed polygon.
    This program finds the rotation angles of each edge of the convex polygon,
    then tests the area of a bounding box aligned with the unique angles in
    90 degrees of the 1st Quadrant.
    Returns the

    Tested with Python 2.6.5 on Ubuntu 10.04.4 (original version)
    Results verified using Matlab

    Copyright (c) 2013, David Butterworth, University of Queensland
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright
          notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.
        * Neither the name of the Willow Garage, Inc. nor the names of its
          contributors may be used to endorse or promote products derived from
          this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
    LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
    CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
    SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
    INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
    CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
    POSSIBILITY OF SUCH DAMAGE.
    */

    if (polygon.outer().empty())
        return;
    BoostPolygon convex_hull;
    bg::convex_hull(polygon, convex_hull);

    //cout << "Convex hull: " << bg::wkt<BoostPolygon2D>(convex_hull) << endl;

    //# Compute edges (x2-x1,y2-y1)
    std::vector<BoostPoint> edges;
    auto convex_hull_outer = convex_hull.outer();
    for (long i=0; i < long(convex_hull_outer.size())-1; ++i) {
        BoostPoint p1      = convex_hull_outer.at(i);
        BoostPoint p2      = convex_hull_outer.at(i+1);
        double edge_x   = p2.get<0>() - p1.get<0>();
        double edge_y   = p2.get<1>() - p1.get<1>();
        edges.push_back(BoostPoint{edge_x, edge_y});
    }

//    cout << "Edges: ";
//    for (auto e : edges)
//        cout << e.get<0>() << " " << e.get<1>() << ",";
//    cout << endl;

    // Calculate unique edge angles  atan2(y/x)
    double angle_scale = 1e3;
    std::set<long> angles_long;
    for (auto vertex : edges) {
        double angle = std::fmod(atan2(vertex.get<1>(), vertex.get<0>()), M_PI / 2);
        angle = angle < 0 ? angle + M_PI / 2 : angle; // want strictly positive answers
        angles_long.insert(long(round(angle*angle_scale)));
    }
    std::vector<double> edge_angles;
    for (auto a : angles_long)
        edge_angles.push_back(double(a)/angle_scale);


//    cout << "Unique angles: ";
//    for (auto e : edge_angles)
//        cout << e*180/M_PI << ",";
//    cout << endl;

    double min_area = std::numeric_limits<double>::infinity();
    // Test each angle to find bounding box with smallest area
    // print "Testing", len(edge_angles), "possible rotations for bounding box... \n"
    for (double angle : edge_angles){

        trans::rotate_transformer<bg::degree, double, 2, 2> rotate(angle*180/M_PI);
        BoostPolygon hull_rotated;
        bg::transform(convex_hull, hull_rotated, rotate);
        //cout << "Convex hull rotated: " << bg::wkt<BoostPolygon2D>(hull_rotated) << endl;

        bg::model::box<BoostPoint> box;
        bg::envelope(hull_rotated, box);
//        cout << "Bounding box: " << bg::wkt<bg::model::box<BoostPoint2D>>(box) << endl;

        //# print "Rotated hull points are \n", rot_points
        BoostPoint min_corner = box.min_corner();
        BoostPoint max_corner = box.max_corner();
        double min_x = min_corner.get<0>();
        double max_x = max_corner.get<0>();
        double min_y = min_corner.get<1>();
        double max_y = max_corner.get<1>();
//        cout << "min_x: " << min_x << endl;
//        cout << "max_x: " << max_x << endl;
//        cout << "min_y: " << min_y << endl;
//        cout << "max_y: " << max_y << endl;

        // Calculate height/width/area of this bounding rectangle
        double width    = max_x - min_x;
        double height   = max_y - min_y;
        double area     = width * height;
//        cout << "Width: " << width << endl;
//        cout << "Height: " << height << endl;
//        cout << "area: " << area << endl;
//        cout << "angle: " << angle*180/M_PI << endl;

        // Store the smallest rect found first (a simple convex hull might have 2 answers with same area)
        if (area < min_area){
            min_area = area;
            minBBox.angle  = angle;
            minBBox.width  = width;
            minBBox.height = height;

            minBBox.corners.clear();
            minBBox.corners.outer().push_back(BoostPoint{min_x, min_y});
            minBBox.corners.outer().push_back(BoostPoint{min_x, max_y});
            minBBox.corners.outer().push_back(BoostPoint{max_x, max_y});
            minBBox.corners.outer().push_back(BoostPoint{max_x, min_y});
            minBBox.corners.outer().push_back(BoostPoint{min_x, min_y});
        }
        //cout << endl << endl;
    }


    // Transform corners of minimal bounding box.
    trans::rotate_transformer<bg::degree, double, 2, 2> rotate(-minBBox.angle*180/M_PI);
    BoostPolygon rotated_polygon;
    bg::transform(minBBox.corners, rotated_polygon, rotate);
    minBBox.corners = rotated_polygon;
}

void toBoost(const Point2D &point, BoostPoint &boost_point)
{
    boost_point.set<0>(point[0]);
    boost_point.set<1>(point[1]);
}

void fromBoost(const BoostPoint &boost_point, Point2D &point)
{
    point[0] = boost_point.get<0>();
    point[1] = boost_point.get<1>();
}

void toBoost(const Point2DList &point_list, BoostPolygon &boost_polygon)
{
    for (auto vertex : point_list) {
        BoostPoint boost_vertex;
        toBoost(vertex, boost_vertex);
        boost_polygon.outer().push_back(boost_vertex);
    }
    bg::correct(boost_polygon);
}

void fromBoost(const BoostPolygon &boost_polygon, Point2DList &point_list)
{
    for (auto boost_vertex : boost_polygon.outer()) {
        Point2D vertex;
        fromBoost(boost_vertex, vertex);
        point_list.push_back(vertex);
    }
}

void rotateDeg(const Point2DList &point_list, Point2DList &rotated_point_list, double degree)
{
    trans::rotate_transformer<bg::degree, double, 2, 2> rotate(degree);
    BoostPolygon boost_polygon;
    toBoost(point_list, boost_polygon);
    BoostPolygon rotated_polygon;
    bg::transform(boost_polygon, rotated_polygon, rotate);
    fromBoost(rotated_polygon, rotated_point_list);
}

void rotateRad(const Point2DList &point_list, Point2DList &rotated_point_list, double rad)
{
    rotateDeg(point_list, rotated_point_list, rad*180/M_PI);
}

bool isClockwise(const Point2DList &point_list)
{
    double orientaion = 0;
    double len = point_list.size();
    for (long i=0; i < len-1; ++i){
        Point2D v1 = point_list[i];
        Point2D v2 = point_list[i+1];
        orientaion += (v2[0]-v1[0])*(v2[1]+v1[1]);
    }
    Point2D v1 = point_list[len-1];
    Point2D v2 = point_list[0];
    orientaion += (v2[0]-v1[0])*(v2[1]+v1[1]);

    return orientaion > 0 ? true : false;
}

void offsetPolygon(const BoostPolygon &polygon, BoostPolygon &polygonOffset, double offset)
{
    bg::strategy::buffer::distance_symmetric<double> distance_strategy(offset);
    bg::strategy::buffer::join_miter join_strategy(3);
    bg::strategy::buffer::end_flat end_strategy;
    bg::strategy::buffer::point_square point_strategy;
    bg::strategy::buffer::side_straight side_strategy;


    bg::model::multi_polygon<BoostPolygon> result;

    bg::buffer(polygon, result, distance_strategy, side_strategy, join_strategy, end_strategy, point_strategy);

    polygonOffset = result[0];

}


void graphFromPolygon(const BoostPolygon &polygon, const BoostLineString &vertices, Matrix<double> &graph)
{
    size_t n = graph.getN();

    for (size_t i=0; i < n; ++i) {
        BoostPoint v1 = vertices[i];
        for (size_t j=i+1; j < n; ++j){
            BoostPoint v2 = vertices[j];
            BoostLineString path{v1, v2};

            double distance = 0;
            if (!bg::within(path, polygon))
                distance = std::numeric_limits<double>::infinity();
            else
                distance = bg::length(path);

            graph.set(i, j, distance);
            graph.set(j, i, distance);
        }
    }

}

bool dijkstraAlgorithm(const size_t numElements,
                       size_t startIndex,
                       size_t endIndex,
                       std::vector<size_t> &elementPath,
                       std::function<double (const size_t, const size_t)> distanceDij)
{
    if (    numElements < 0
         || startIndex < 0
         || endIndex < 0
         || startIndex >= numElements
         || endIndex >= numElements
         || endIndex == startIndex) {
        return false;
    }
    // Node struct
    // predecessorIndex is the index of the predecessor node (nodeList[predecessorIndex])
    // distance is the distance between the node and the start node
    // node number is stored by the position in nodeList
    struct Node{
        int predecessorIndex = -1;
        double distance = std::numeric_limits<double>::infinity();
    };

    // The list with all Nodes (elements)
    std::vector<Node> nodeList(numElements);
    // This list will be initalized with indices referring to the elements of nodeList.
    // Elements will be successively remove during the execution of the Dijkstra Algorithm.
    std::vector<size_t> workingSet(numElements);

    //append elements to node list
    for (size_t i = 0; i < numElements; ++i) workingSet[i] = i;


    nodeList[startIndex].distance = 0;

    // Dijkstra Algorithm
    // https://de.wikipedia.org/wiki/Dijkstra-Algorithmus
    while (workingSet.size() > 0) {
        // serach Node with minimal distance
        double minDist = std::numeric_limits<double>::infinity();
        int minDistIndex_WS = -1; // WS = workinSet
        for (size_t i = 0; i < workingSet.size(); ++i) {
            const int nodeIndex = workingSet.at(i);
            const double dist = nodeList.at(nodeIndex).distance;
            if (dist < minDist) {
                minDist = dist;
                minDistIndex_WS = i;
            }
        }
        if (minDistIndex_WS == -1)
                return false;

        size_t indexU_NL = workingSet.at(minDistIndex_WS); // NL = nodeList
        workingSet.erase(workingSet.begin()+minDistIndex_WS);
        if (indexU_NL == endIndex) // shortest path found
            break;

        const double distanceU = nodeList.at(indexU_NL).distance;
        //update distance
        for (size_t i = 0; i < workingSet.size(); ++i) {
            int indexV_NL = workingSet[i]; // NL = nodeList
            Node* v = &nodeList[indexV_NL];
            double dist = distanceDij(indexU_NL, indexV_NL);
            // is ther an alternative path which is shorter?
            double alternative = distanceU + dist;
            if (alternative < v->distance)  {
                v->distance         = alternative;
                v->predecessorIndex = indexU_NL;
            }
        }

    }
    // end Djikstra Algorithm


    // reverse assemble path
    int e = endIndex;
    while (1) {
        if (e == -1) {
            if (elementPath[0] == startIndex)// check if starting point was reached
                break;
            return false;
        }
        elementPath.insert(elementPath.begin(), e);

        //Update Node
        e = nodeList[e].predecessorIndex;

    }
    return true;
}

void toDistanceMatrix(Matrix<double> &graph)
{
    size_t n = graph.getN();

    auto distance = [graph](size_t i, size_t j){
        return graph.get(i,j);
    };


    std::vector<size_t> path;
    for (size_t i=0; i < n; ++i) {
        for (size_t j=i+1; j < n; ++j){
            double d = graph.get(i,j);
            if (!std::isinf(d))
                continue;
            path.clear();
            bool ret = dijkstraAlgorithm(n, i, j, path, distance);
            assert(ret);
            (void)ret;
//            cout << "(" << i << "," << j << ") d: " << d << endl;
//            cout << "Path size: " << path.size() << endl;
//            for (auto idx : path)
//                cout << idx << " ";
//            cout << endl;

            d = 0;
            for (long k=0; k < long(path.size())-1; ++k) {
                size_t idx0 = path[k];
                size_t idx1 = path[k+1];
                double d0 = graph.get(idx0, idx1);
                assert(std::isinf(d0) == false);
                d += d0;
            }

            graph.set(i, j, d);
            graph.set(j, i, d);
        }
    }
}

void shortestPathFromGraph(const Matrix<double> &graph, size_t startIndex, size_t endIndex, std::vector<size_t> &pathIdx)
{

    if (!std::isinf(graph.get(startIndex, endIndex))){
        pathIdx.push_back(startIndex);
        pathIdx.push_back(endIndex);
    } else {
        auto distance = [graph](size_t i, size_t j){
            return graph.get(i, j);
        };
        bool ret = dijkstraAlgorithm(graph.getN(), startIndex, endIndex, pathIdx, distance);
        assert(ret);
        (void)ret;
    }

}

} // end namespace snake_geometry
